<?php

declare(strict_types=1);

namespace DigiComp\PhpCodesniffer\DigiComp\Sniffs\Annotations;

use DigiComp\PhpCodesniffer\DigiComp\Sniffs\MemberVarUtility;
use PHP_CodeSniffer\Files\File;
use PHP_CodeSniffer\Sniffs\AbstractVariableSniff;

class VarIsLastTagOnPropertySniff extends AbstractVariableSniff
{
    /**
     * @inheritDoc
     */
    protected function processVariable(File $phpcsFile, $stackPtr)
    {
    }

    /**
     * @inheritDoc
     */
    protected function processVariableInString(File $phpcsFile, $stackPtr)
    {
    }

    /**
     * @inheritDoc
     */
    protected function processMemberVar(File $phpcsFile, $stackPtr)
    {
        try {
            $propertyInfo = $phpcsFile->getMemberProperties($stackPtr);
            if ($propertyInfo === []) {
                return;
            }
        } catch (\Exception $e) {
            // Turns out not to be a property after all.
            return;
        }
        $tokens = $phpcsFile->getTokens();
        $ignore = [
            \T_PUBLIC => \T_PUBLIC,
            \T_PRIVATE => \T_PRIVATE,
            \T_PROTECTED => \T_PROTECTED,
            \T_VAR => \T_VAR,
            \T_STATIC => \T_STATIC,
            \T_WHITESPACE => \T_WHITESPACE,
            \T_STRING => \T_STRING,
            \T_NS_SEPARATOR => \T_NS_SEPARATOR,
            \T_NULLABLE => \T_NULLABLE,
        ];

        $commentEnd = MemberVarUtility::findCommentEndWithoutAttributes($tokens, $ignore, $stackPtr);
        if (
            $commentEnd === false
            || ($tokens[$commentEnd]['code'] !== \T_DOC_COMMENT_CLOSE_TAG
                && $tokens[$commentEnd]['code'] !== \T_COMMENT)
        ) {
            // Obviously, there is NO comment at all, we are done
            return;
        }

        $commentStart = $tokens[$commentEnd]['comment_opener'];

        $varToken = $otherToken = null;
        foreach ($tokens[$commentStart]['comment_tags'] as $tag) {
            if ($tokens[$tag]['content'] === '@var') {
                $varToken = $tag;
            } else {
                if ($otherToken === null || $tag > $otherToken) {
                    $otherToken = $tag;
                }
            }
        }
        if ($varToken !== null && $otherToken > $varToken) {
            $fix = $phpcsFile->addFixableError('@var should be last property annotation', $varToken, 'VarNotLast');
            if ($fix) {
                $this->moveVarTagToEndOfDoctype(
                    $varToken,
                    $tokens[$commentStart]['comment_tags'],
                    $phpcsFile,
                    $commentEnd
                );
            }
        }
    }

    /**
     * @param int $varToken
     * @param array $commentTags
     * @param File $phpcsFile
     * @param int $commentEnd
     */
    protected function moveVarTagToEndOfDoctype(
        int $varToken,
        array $commentTags,
        File $phpcsFile,
        int $commentEnd
    ): void {
        $nextTokenIndex = \array_search($varToken, $commentTags) + 1;
        $nextToken = $commentTags[$nextTokenIndex];
        // subtracting four tokens (whitespace + "*" + whitespace + newline) as we already know, we are not last
        $varContent = $phpcsFile->getTokensAsString($varToken - 4, $nextToken - $varToken);
        $phpcsFile->fixer->beginChangeset();
        $phpcsFile->fixer->addContentBefore($commentEnd - 2, $varContent);
        for ($tokenToDelete = $varToken; $tokenToDelete < $nextToken; $tokenToDelete++) {
            $phpcsFile->fixer->replaceToken($tokenToDelete, '');
        }
        $phpcsFile->fixer->endChangeset();
    }
}
